/* Include file for curve.c
 * Defines the Point class, and the Fp2 class
 * Ben Lynn
 */
/*
Copyright (C) 2001 Benjamin Lynn (blynn@cs.stanford.edu)

This file is part of the Stanford IBE system.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef CURVE_H
#define CURVE_H

#include <openssl/bn.h>
#include "fp2.h"

#ifdef __cplusplus
extern "C" {
#endif

void curve_init_prime(mpz_t prime, mpz_t qprime);
//initializes system parameters
//should be first library function called
//not thread-safe

void curve_shutdown(void);

struct point_s {
    fp2_t x, y;
    int infinity;
};

typedef struct point_s point_t[1];
typedef struct point_s *point_ptr;

void point_init(point_ptr P);
//allocates memory for a point

size_t point_out_str(FILE *stream, int base, point_ptr P);

void point_clear(point_ptr P);
//deallocates memory for a point

void point_set(point_ptr src, point_ptr dst);
//src = dst

void point_set_O(point_ptr P);
//P = O

int point_equal(point_t P, point_t Q);
//P == Q?

void point_add(point_ptr R, point_ptr P, point_ptr Q);
//R = P + Q

void x_from_y(mpz_t x, mpz_t y);
//x = cuberoot(y^2 - 1)

void mympz_randomp(mpz_t x);
//x = random in {0, ..., p - 1}

void mympz_randomm(mpz_t x, mpz_t limit);
//x = random in {0, ..., limit - 1}

void fp2_random(fp2_ptr r);
//r = random element of F_p^2

void general_point_random(point_ptr P);
//P = random point on E/F_p^2

void point_random(point_ptr P);
//P = random point

void weil_pairing(fp2_ptr res, point_ptr P, point_ptr Q);
//res = e(P, Q) where e is the Weil pairing

void tate_pairing(fp2_ptr res, point_ptr P, point_ptr Q);
// res = e(P,Q) where e is the Tate pairing

void point_mul(point_ptr R, mpz_t n, point_ptr P);
//R = nP
//P must lie on E/F_p, n must be positive

void general_point_mul(point_t R, mpz_t n, point_t P);
//R = nP
//can handle P on E/F_p^2, any integer n

void mympz_set_bn(mpz_t z, BIGNUM *bn);
//z = bn
void mympz_get_bn(mpz_t z, BIGNUM *bn);
//bn = z

void point_mul_preprocess(point_ptr P);
void point_mul_postprocess(point_ptr res, mpz_t n);

struct miller_cache_s {
    mpz_t *numa, *numc, *denomc;
    mpz_t denomsb;
    mpz_t denoms1;
    mpz_t numl1a, numl1c, denoml1c;
    mpz_t numl2c;
};

typedef struct miller_cache_s miller_cache_t[1];

void miller_cache_init(miller_cache_t mc);

void tate_preprocess(miller_cache_t mc, point_ptr P);
void miller_postprocess(fp2_ptr res, miller_cache_t mc, point_ptr Q);
void tate_postprocess(fp2_ptr res, miller_cache_t mc, point_ptr Q);

#ifdef __cplusplus
}
#endif

#endif //CURVE_H
