/* F_p^2 routines
 * Ben Lynn
 */
/*
Copyright (C) 2001 Benjamin Lynn (blynn@cs.stanford.edu)

This file is part of the Stanford IBE system.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "fp2.h"

enum {
    //constants for sliding window algorithms
    windowsize = 5,
    windowsizepower = 15	    //this is 2^(windowsize - 1) - 1
};

static mpz_t modulus;	//holds p
//we represent elements of F_p^2 with
//a + sqrt(-1)b where a, b lie in F_p

//why aren't there gmp functions for arithmetic modulo p?
//these are inefficient?
static void reduce(mpz_t x)
{
    mpz_mod(x, x, modulus);
}

static void sqr_mod(mpz_t x, mpz_t a)
{
    mpz_mul(x, a, a);
    reduce(x);
}

static void mul_mod(mpz_t x, mpz_t a, mpz_t b)
{
    mpz_mul(x, a, b);
    reduce(x);
}

static void add_mod(mpz_t x, mpz_t a, mpz_t b)
{
    mpz_add(x, a, b);
    if (mpz_cmp(x, modulus) >= 0) mpz_sub(x, x, modulus);
    /*
    mpz_add(x, a, b);
    reduce(x);
    */
}

static void neg_mod(mpz_t x, mpz_t b)
{
    if (mpz_sgn(b)) {
	mpz_sub(x, modulus, b);
    }
    /*
    mpz_sub(x, a, b);
    reduce(x);
    */
}

static void sub_mod(mpz_t x, mpz_t a, mpz_t b)
{
    if (mpz_cmp(a, b) >= 0) {
	mpz_sub(x, a, b);
    } else {
	mpz_sub(x, a, b);
	mpz_add(x, x, modulus);
    }
    /*
    mpz_sub(x, a, b);
    reduce(x);
    */
}

static void inv_mod(mpz_t x, mpz_t a)
{
    mpz_invert(x, a, modulus);
}

void zp_add(mpz_t x, mpz_t a, mpz_t b)
//x = a + b mod p
{
    add_mod(x, a, b);
}

void zp_sub(mpz_t x, mpz_t a, mpz_t b)
//x = a - b mod p
{
    sub_mod(x, a, b);
}

void zp_neg(mpz_t x, mpz_t b)
//x = -b mod p
{
    neg_mod(x, b);
}

void zp_mul(mpz_t x, mpz_t a, mpz_t b)
//x = a * b mod p
{
    mul_mod(x, a, b);
}

void zp_invert(mpz_t x, mpz_t a)
//x = 1/a mod p
{
    inv_mod(x, a);
}

void fp2_init_modulus(mpz_t m)
//should be first call to library
//sets p in F_p^2 to modulus
{
    mpz_init(modulus);
    mpz_set(modulus, m);
}

void fp2_shutdown(void)
//shutdown fp2 library
{
    mpz_clear(modulus);
}

void fp2_init(fp2_ptr x)
//call before using x
//allocates space for x
{
    mpz_init(x->a);
    mpz_init(x->b);
}

void fp2_clear(fp2_ptr x)
//deallocates space for x
{
    mpz_clear(x->a);
    mpz_clear(x->b);
}

void fp2_set(fp2_ptr x, fp2_ptr y)
//assignment
//x = y
{
    mpz_set(x->a, y->a);
    mpz_set(x->b, y->b);
}

void fp2_set_0(fp2_ptr x)
//x = 0
{
    mpz_set_ui(x->a, 0);
    mpz_set_ui(x->b, 0);
}

void fp2_set_1(fp2_ptr x)
//x = 1
{
    mpz_set_ui(x->a, 1);
    mpz_set_ui(x->b, 0);
}

void fp2_set_cbrt_unity(fp2_ptr x)
{
    //Warning: depends on representation of F_p^2
    //Z_p[sqrt(-1)] version:
    mpz_t p1on4;

    mpz_set(x->a, modulus);
    mpz_sub_ui(x->a, x->a, 1);
    mpz_div_ui(x->a, x->a, 2);

    mpz_init_set(p1on4, modulus);
    mpz_add_ui(p1on4, p1on4, 1);
    mpz_div_ui(p1on4, p1on4, 4);
    mpz_set_ui(x->b, 3);
    mpz_powm(x->b, x->b, p1on4, modulus);
    //TODO: inefficient way of halving
    mpz_mul(x->b, x->b, x->a);
    mpz_mod(x->b, x->b, modulus);
    mpz_clear(p1on4);

    //Z_p[sqrt(-3)] version:
    //mpz_set(x->a, modulus);
    //mpz_sub_ui(x->a, x->a, 1);
    //mpz_div_ui(x->a, x->a, 2);
    //mpz_set(x->b, x->a);
}

void fp2_init_set(fp2_ptr x, fp2_ptr y)
//for convenience
//combines fp2_init, fp2_set
{
    mpz_init(x->a);
    mpz_init(x->b);
    mpz_set(x->a, y->a);
    mpz_set(x->b, y->b);
}

void fp2_neg(fp2_ptr x, fp2_ptr a)
//x = -a
{
    if (mpz_sgn(a->a)) {
	mpz_sub(x->a, modulus, a->a);
    }
    if (mpz_sgn(a->b)) {
	mpz_sub(x->b, modulus, a->b);
    }
}

void fp2_add(fp2_ptr x, fp2_ptr a, fp2_ptr b)
//x = a + b
{
    add_mod(x->a, a->a, b->a);
    add_mod(x->b, a->b, b->b);
}

void fp2_sub(fp2_ptr x, fp2_ptr a, fp2_ptr b)
//x = a - b
{
    sub_mod(x->a, a->a, b->a);
    sub_mod(x->b, a->b, b->b);
}

void fp2_sub_ui(fp2_ptr x, fp2_ptr a, unsigned long b)
//x = a - b
{
    mpz_sub_ui(x->a, a->a, b);
    reduce(x->a);
}

void fp2_sqr(fp2_ptr x, fp2_ptr a)
//x = a * a
{
    mpz_t t0, t1; //temp variables
    mpz_init(t0); mpz_init(t1);

    //sqr_mod(t0, a->a);
    //sqr_mod(t1, a->b);
    //sub_mod(t0, t0, t1);

    /*
    sub_mod(t0, a->a, a->b);
    add_mod(t1, a->a, a->b);
    mul_mod(t0, t0, t1);
    mul_mod(t1, a->a, a->b);
    add_mod(x->b, t1, t1);
    mpz_set(x->a, t0);
    */
    mpz_sub(t0, a->a, a->b);
    mpz_add(t1, a->a, a->b);
    mpz_mul(t0, t0, t1);
    mpz_mul(t1, a->a, a->b);
    mpz_mul_2exp(x->b, t1, 1);
    reduce(x->b);
    mpz_mod(x->a, t0, modulus);

    mpz_clear(t0); mpz_clear(t1);
}

void fp2_mul_mpz(fp2_ptr x, fp2_ptr a, mpz_ptr b)
//x = a * b
{
    mul_mod(x->a, a->a, b);
    mul_mod(x->b, a->b, b);
}

void fp2_mul(fp2_ptr x, fp2_ptr a, fp2_ptr b)
//x = a * b
{
    mpz_t t0, t1, t2; //temp variables
    mpz_init(t0); mpz_init(t1); mpz_init(t2);

    /* This is slower
    mul_mod(t0, a->a, b->a);
    mul_mod(t1, a->b, b->b);
    sub_mod(t0, t0, t1);
    mul_mod(t1, a->a, b->b);
    mul_mod(t2, a->b, b->a);
    add_mod(x->b, t1, t2);
    mpz_set(x->a, t0);
    */
    mpz_mul(t0, a->a, b->a);
    mpz_mul(t1, a->b, b->b);
    mpz_sub(t0, t0, t1);
    mpz_mul(t1, a->a, b->b);
    mpz_mul(t2, a->b, b->a);
    mpz_add(x->b, t1, t2);
    reduce(x->b);
    mpz_mod(x->a, t0, modulus);

    mpz_clear(t0); mpz_clear(t1); mpz_clear(t2);
}

void fp2_inv(fp2_ptr x, fp2_ptr b)
//x = 1 / b
{
    mpz_t t0, t1; //temp variables
    mpz_t t2;
    mpz_init(t2);
    mpz_init(t0); mpz_init(t1);

    /*
    mul_mod(t0, b->a, b->a);
    mul_mod(t1, b->b, b->b);
    add_mod(t0, t0, t1);
    inv_mod(t1, t0);

    mul_mod(t0, t1, b->b);
    neg_mod(x->b, t0);
    mul_mod(x->a, t1, b->a);
    */
    mpz_mul(t0, b->a, b->a);
    mpz_mul(t1, b->b, b->b);
    mpz_add(t0, t0, t1);
    mpz_invert(t1, t0, modulus);

    mpz_mul(t0, t1, b->b);
    mpz_neg(x->b, t0);
    reduce(x->b);
    mpz_mul(x->a, t1, b->a);
    reduce(x->a);

    mpz_clear(t0); mpz_clear(t1);
    mpz_clear(t2);
}

void fp2_div(fp2_ptr x, fp2_ptr a, fp2_ptr b)
//x = a / b
{
    mpz_t t0, t1, t2, t3; //temp variables
    mpz_init(t0); mpz_init(t1); mpz_init(t2); mpz_init(t3);

    /*
    mul_mod(t0, b->a, b->a);
    mul_mod(t1, b->b, b->b);
    add_mod(t0, t0, t1);
    inv_mod(t1, t0);

    mul_mod(t0, a->a, b->a);
    mul_mod(t2, a->b, b->b);
    add_mod(t0, t0, t2);
    mul_mod(t2, a->b, b->a);
    mul_mod(t3, a->a, b->b);
    sub_mod(t2, t2, t3);
    mul_mod(x->b, t1, t2);
    mul_mod(x->a, t1, t0);
    */
    mpz_mul(t0, b->a, b->a);
    mpz_mul(t1, b->b, b->b);
    mpz_add(t0, t0, t1);
    mpz_invert(t1, t0, modulus);

    mpz_mul(t0, a->a, b->a);
    mpz_mul(t2, a->b, b->b);
    mpz_add(t0, t0, t2);
    mpz_mul(t2, a->b, b->a);
    mpz_mul(t3, a->a, b->b);
    mpz_sub(t2, t2, t3);
    mpz_mul(x->b, t1, t2);
    reduce(x->b);
    mpz_mul(x->a, t1, t0);
    reduce(x->a);

    mpz_clear(t0); mpz_clear(t1); mpz_clear(t2); mpz_clear(t3);
}

void fp2_pow(fp2_ptr res, fp2_ptr x, mpz_t n)
//exponentiation
//res = x^n
{
    int i, j, k, l, m;
    mpz_t t0, t1; //temp variables

    //use sliding-window method
    fp2_t g[2*windowsizepower+2];
    fp2_init_set(g[1], x);
    fp2_init(g[2]);
    fp2_mul(g[2], x, x);

    mpz_init(t0); mpz_init(t1);

    for (i=1; i<=windowsizepower; i++) {
	j = 2 * i + 1;
	fp2_init(g[j]);
	fp2_mul(g[j], g[j - 2], g[2]);
    }
    m = mpz_sizeinbase(n, 2) - 1;

    fp2_set_1(res);

    while(m>=0) {
	if (!mpz_tstbit(n, m)) {
	    { //fp2_sqr(res,res);
		mpz_sub(t0, res->a, res->b);
		mpz_add(t1, res->a, res->b);
		mpz_mul(t0, t0, t1);
		mpz_mul(t1, res->a, res->b);
		mpz_mul_2exp(res->b, t1, 1);
		reduce(res->b);
		mpz_mod(res->a, t0, modulus);
	    }
	    m--;
	} else {
	    l = m - windowsize + 1;
	    if (l < 0) l = 0;
	    l = mpz_scan1(n, l);
	    j = 1;
	    { //fp2_sqr(res,res);
		mpz_sub(t0, res->a, res->b);
		mpz_add(t1, res->a, res->b);
		mpz_mul(t0, t0, t1);
		mpz_mul(t1, res->a, res->b);
		mpz_mul_2exp(res->b, t1, 1);
		reduce(res->b);
		mpz_mod(res->a, t0, modulus);
	    }
	    for (k = m - 1; k>=l; k--) {
		j = j << 1;
		if (mpz_tstbit(n, k)) j++;
		{ //fp2_sqr(res,res);
		    mpz_sub(t0, res->a, res->b);
		    mpz_add(t1, res->a, res->b);
		    mpz_mul(t0, t0, t1);
		    mpz_mul(t1, res->a, res->b);
		    mpz_mul_2exp(res->b, t1, 1);
		    reduce(res->b);
		    mpz_mod(res->a, t0, modulus);
		}
	    }
	    fp2_mul(res, res, g[j]);
	    m = l-1;
	}
    }

    fp2_clear(g[2]);
    for (i=0; i<=windowsizepower; i++) {
	j = 2 * i + 1;
	fp2_clear(g[j]);
    }

    mpz_clear(t0); mpz_clear(t1);

    /* simple method:
    int m = mpz_sizeinbase(n, 2) - 1;
    fp2_t x1;

    fp2_init_set(x1, x);

    for(;;) {
	if (mpz_tstbit(n, m)) {
	    fp2_mul(res, res, x1);
	}
	if (m == 0) break;
	fp2_mul(res, res, res);
	m--;
    }
    */
}

int fp2_is_0(fp2_ptr x)
// x == 0?
{
    return (!(mpz_size(x->a) || mpz_size(x->b)));
}

int fp2_equal(fp2_ptr x, fp2_ptr y)
// x == y?
{
    return (!(mpz_cmp(x->a, y->a) || mpz_cmp(x->b, y->b)));
}

size_t fp2_out_str(FILE *stream, int base, fp2_ptr x)
//output x on stream in given base
{
    FILE *fp;
    size_t s, status;

    if (!stream) fp = stdout;
    else fp = stream;
    status = fprintf(fp, "[");
    if (status < 0) return status;
    s = status;
    status = mpz_out_str(fp, base, x->a);
    if (status < 0) return status;
    s += status;
    status = fprintf(fp, " ");
    if (status < 0) return status;
    s += status;
    status = mpz_out_str(fp, base, x->b);
    if (status < 0) return status;
    s += status;
    status = fprintf(fp, "]");
    if (status < 0) return status;
    s += status;
    return s;
}
