/* Operations in F_p^2
 * header file
 * Ben Lynn
 */
/*
Copyright (C) 2001 Benjamin Lynn (blynn@cs.stanford.edu)

This file is part of the Stanford IBE system.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef FP2_H
#define FP2_H

#include <stdio.h>
#include <gmp.h>

#ifdef __cplusplus
extern "C" {
#endif

//for F_p^2 we use the representation a + sqrt(-nqr)b
struct fp2_s {
    mpz_t a, b;
};

typedef struct fp2_s fp2_t[1];
typedef struct fp2_s *fp2_ptr;

//these assume the arguments are in {0,...,p-1}

void zp_add(mpz_t x, mpz_t a, mpz_t b);
//x = a + b mod p

void zp_neg(mpz_t x, mpz_t b);
//x = -b mod p

void zp_sub(mpz_t x, mpz_t a, mpz_t b);
//x = a - b mod p

void zp_mul(mpz_t x, mpz_t a, mpz_t b);
//x = a * b mod p

void zp_invert(mpz_t x, mpz_t a);
//x = 1/a mod p
//mimic GMP's mpz style

void fp2_init_modulus(mpz_t m);
//should be first call to library
//sets p in F_p^2 to modulus

void fp2_shutdown(void);
//shutdown fp2 library


void fp2_init(fp2_ptr x);
//call before using x
//allocates space for x

void fp2_clear(fp2_ptr x);
//deallocates space for x

void fp2_set(fp2_ptr x, fp2_ptr y);
//x = y

void fp2_set_0(fp2_ptr x);
//x = 0

void fp2_set_1(fp2_ptr x);
//x = 1

void fp2_set_cbrt_unity(fp2_ptr x);
//x = nonunit cube root of unity

void fp2_init_set(fp2_ptr x, fp2_ptr y);
//for convenience
//combines fp2_init, fp2_set

void fp2_neg(fp2_ptr x, fp2_ptr a);
//x = -a

void fp2_add(fp2_ptr x, fp2_ptr a, fp2_ptr b);
//x = a + b

void fp2_sub(fp2_ptr x, fp2_ptr a, fp2_ptr b);
//x = a - b

void fp2_sub_ui(fp2_ptr x, fp2_ptr a, unsigned long b);
//x = a - b

void fp2_sqr(fp2_ptr x, fp2_ptr a);
//x = a * a

void fp2_mul_mpz(fp2_ptr x, fp2_ptr a, mpz_ptr b);
//x = a * b

void fp2_mul(fp2_ptr x, fp2_ptr a, fp2_ptr b);
//x = a * b

void fp2_inv(fp2_ptr x, fp2_ptr b);
//x = 1 / b

void fp2_div(fp2_ptr x, fp2_ptr a, fp2_ptr b);
//x = a / b

void fp2_pow(fp2_ptr res, fp2_ptr x, mpz_t n);
//res = x^n

int fp2_equal(fp2_ptr x, fp2_ptr y);
//returns true if x equals y, false otherwise

int fp2_is_0(fp2_ptr x);
//returns true if x equals 0, false otherwise

size_t fp2_out_str(FILE *stream, int base, fp2_ptr x);
//output x on stream in base base

#ifdef __cplusplus
}
#endif

#endif //FP2_H
