/* IBE libary function prototypes
 * names are almost self-explanatory
 * Ben Lynn
 */
/*
Copyright (C) 2001 Benjamin Lynn (blynn@cs.stanford.edu)

This file is part of the Stanford IBE system.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef IBE_H
#define IBE_H

#ifndef __KERNEL__
#include <stdio.h>
#endif

struct byte_string_s {
    unsigned char *data;
    int len;
};

typedef struct byte_string_s byte_string_t[1];
typedef struct byte_string_s *byte_string_ptr;

void byte_string_init(byte_string_t bs, int n);
void byte_string_reinit(byte_string_t bs, int n);
void byte_string_init_set(byte_string_t bs, const char *s);
#ifndef __KERNEL__
void byte_string_fprintf(FILE *fp, byte_string_t bs, char *format);
#endif
void byte_string_printf(byte_string_t bs, char *format);
void byte_string_clear(byte_string_t bs);

void IBE_init(void);
//initialize library

//The four basic operations; see paper

void IBE_setup(int k, int qk, char *system);
//generate system parameters

void IBE_extract(byte_string_t key, const char *id);
//extract private key corresponding to given ID
//shouldn't normally be used because the master shouldn't be known
//by only one server

//These are not very well written. They can only encrypt/decrypt things
//that fit in memory

int IBE_encrypt(byte_string_t U, byte_string_t *Varray, byte_string_t W,
	byte_string_t M, char **idarray);
//idarray is a null-terminated array of ID's
//<U,V,W> = encrypt(M)
//(U and W are the same for every id)

int IBE_decrypt(byte_string_t M, byte_string_t U, byte_string_t V,
	byte_string_t W, byte_string_t key);
//M = decrypt(<U,V,W>)

int IBE_authenticated_decrypt(byte_string_t M, byte_string_t U, byte_string_t V,
	byte_string_t W, byte_string_t d, char *sender);

int IBE_authenticated_encrypt(byte_string_t U, byte_string_t *Varray, byte_string_t W,
	byte_string_t M, char **idarray,
	byte_string_t sender_key);

//these are called by encrypt and decrypt
//use these if you want to write your own encrypt/decrypt
//sigma is the key that is actually used to encrypt the plaintext
void IBE_hidesigma(byte_string_t U, byte_string_t *Varray,
    byte_string_t sigma, byte_string_t hash,
    char **idarray);
//encrypt a key sigma (which must be n bits long)
//output in U, Varray
//must be given the hash of the plaintext message,
//and idarray holds the public keys to encrypt for
//requires Varray has same length has idarray

void IBE_revealsigma(byte_string_t sigma,
	byte_string_t U, byte_string_t V, byte_string_t key);
//reveal a key sigma
//given U, V, and a private key
void IBE_revealsigma_preprocess(byte_string_t key);
void IBE_revealsigma_postprocess(byte_string_t sigma,
	byte_string_t U, byte_string_t V);

int IBE_verify(byte_string_t U, byte_string_t sigma, byte_string_t hash);
//verify that the ciphertext is valid
//given U, a key sigma, and the hash of the plaintext


void IBE_split_master(byte_string_t *mshare, int t, int n);
//split the master key into n pieces
//t of them are required to generate a key

int IBE_construct_master(byte_string_t *mshare, int t);
//reconstruct the master from master shares
//(shouldn't normally be used since they should never be in one place)

int IBE_serialize_params(byte_string_t bs);
//put system parameters into a byte_string
int IBE_deserialize_params(byte_string_t bs);
//get system parameters from a byte_string

int IBE_load_master_share(byte_string_t mshare);
//load a master share from a byte_string
void IBE_clear_master_share(byte_string_t mshare);
//free previously loaded master share
int IBE_extract_share(byte_string_t share, const char *id);
//extract a private key share from an ID
//(using previously loaded master share)
int IBE_construct_key(byte_string_t key, byte_string_t *kshare);
//reconstruct a key from key shares

char *IBE_make_id(char *addr, char *subject);
//adds fields to email address to make it a valid ID
int IBE_n(void);
//returns n (see paper)

int IBE_threshold(void);
int IBE_sharecount(void);
char* IBE_system(void);
char* IBE_version(void);

void IBE_get_shared_secret(byte_string_t s, char *id, byte_string_t key);
void IBE_get_shared_secret_preprocess(byte_string_t key);
void IBE_get_shared_secret_postprocess(byte_string_t key, char *id);

void IBE_print_params(void);

#endif //IBE_H
