/* IBE library routines, see ibe.h for usage
 * Ben Lynn
 */

/*
Copyright (C) 2001 Benjamin Lynn (blynn@cs.stanford.edu)

This file is part of the Stanford IBE system.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/bn.h>
#include <string.h>
#include "curve.h"
#include "version.h"
#include "benchmark.h"
#include "ibe.h"

struct params_s {
    char *id;
    char *version;
    mpz_t p, q;
    int n;
    point_t P;
    point_t Ppub;

    int sharet, sharen;
    mpz_t *robustx;
    point_t *robustP;

};

static struct params_s params;
static mpz_t master;
static mpz_t p1onq;
static fp2_t zeta; //cube root of unity
static char *version_string;
static point_t PhiPpub;

static EVP_MD_CTX mdctx;
const static EVP_MD *md;
static EVP_CIPHER* sharecipher;

static mpz_t current_master_share_y;
static int current_master_share_i;

static miller_cache_t cache_ssmc; //shared secret miller cache
static miller_cache_t cache_xQmc; //xQ miller cache
static miller_cache_t cache_Ppmc; //Ppub miller cache

int mympz_sizeinbytes(mpz_t x)
{
    int i, n;
    n = mpz_sizeinbase(x, 2);
    i = n >> 3;
    if (n % 8) i++;

    return i;
}

void mympz_out_raw(mpz_t x, unsigned char *c, int n)
{
    unsigned long l;
    mpz_t z;
    int i = n - 1;

    mpz_init_set(z, x);

    while(i >= 0) {
	l = mpz_get_ui(z);
	c[i] = (unsigned char) l;
	mpz_tdiv_q_2exp(z, z, 8);
	i--;
    }
    mpz_clear(z);
}

void mympz_inp_raw(mpz_t z, const unsigned char* c, int n)
{
    int i = 0;

    mpz_set_ui(z, 0);
    while(i < n) {
	mpz_mul_2exp(z, z, 8);
	mpz_add_ui(z, z, c[i]);
	i++;
    }
}

void byte_string_init(byte_string_t bs, int n)
{
    bs->data = (unsigned char *) malloc(n * sizeof(unsigned char));
    bs->len = n;
}

void byte_string_reinit(byte_string_t bs, int n)
{
    bs->data = (unsigned char *) realloc(bs->data, n * sizeof(unsigned char));
    bs->len = n;
}

void byte_string_fprintf(FILE *fp, byte_string_t bs, char *format)
{
    int i;
    for(i=0; i<bs->len; i++) {
	fprintf(fp, format, bs->data[i]);
    }
}

void byte_string_printf(byte_string_t bs, char *format)
{
    int i;
    for(i=0; i<bs->len; i++) {
	printf(format, bs->data[i]);
    }
}

void byte_string_cat(byte_string_t bs, byte_string_t bs2)
//concatenates bs2 to bs
//automatically clears bs2
{
    int n;
    n = bs->len + bs2->len;
    bs->data = (unsigned char *) realloc(bs->data, n * sizeof(unsigned char));
    memcpy(&bs->data[bs->len], bs2->data, bs2->len);
    bs->len = n;
    byte_string_clear(bs2);
}

void byte_string_init_set(byte_string_t bs, const char *s)
{
    bs->len = strlen(s);
    bs->data = (unsigned char *) malloc(bs->len * sizeof(unsigned char));
    memcpy(bs->data, s, bs->len);
}

void byte_string_copy(byte_string_t bs, byte_string_t src)
{
    bs->len = src->len;
    bs->data = (unsigned char *) malloc(bs->len * sizeof(unsigned char));
    memcpy(bs->data, src->data, bs->len);
}

void byte_string_init_set_mpz(byte_string_t bs, mpz_t x)
{
    int i = mympz_sizeinbytes(x);

    byte_string_init(bs, i + 2);
    bs->data[0] = (unsigned char) (i >> 8);
    bs->data[1] = (unsigned char) i;
    mympz_out_raw(x, &bs->data[2], i);
}

void byte_string_init_set_fp2(byte_string_t bs, fp2_t x)
{
    byte_string_t bsb;
    byte_string_init_set_mpz(bs, x->a);
    byte_string_init_set_mpz(bsb, x->b);
    byte_string_cat(bs, bsb);
}

void byte_string_init_set_point(byte_string_t bs, point_t P)
{
    byte_string_t bsy;
    byte_string_init_set_fp2(bs, P->x);
    byte_string_init_set_fp2(bsy, P->y);
    byte_string_cat(bs, bsy);
}

void byte_string_clear(byte_string_t bs)
{
    free(bs->data);
    bs->len = 0;
}

int mympz_set_data_len(mpz_t z, unsigned char *data, int len)
{
    int n;

    if (len < 2) {
	fprintf(stderr, "bad serialization: len < 2\n");
	return 0;
    }
    n = (data[0] << 8) + data[1];
    if (n + 2 > len) {
	fprintf(stderr, "bad serialization: n + 2 > len\n");
	return 0;
    }
    mympz_inp_raw(z, &data[2], n);
    return n + 2;
}

int mympz_set_byte_string(mpz_t z, byte_string_t bs)
{
    return mympz_set_data_len(z, bs->data, bs->len);
}

int point_set_data_len(point_t P, unsigned char *data, int len)
{
    int i;
    i = mympz_set_data_len(P->x->a, data, len);
    i += mympz_set_data_len(P->x->b, &data[i], len - i);
    i += mympz_set_data_len(P->y->a, &data[i], len - i);
    i += mympz_set_data_len(P->y->b, &data[i], len - i);
    return i;
}

int point_set_byte_string(point_t P, byte_string_t bs)
{
    int i;
    i = point_set_data_len(P, bs->data, bs->len);
    //check i = bs->len
    return i;
}

char *MM_new_string(int n)
{
    char *res;
    res = (char *) malloc(n * sizeof(char));
    return res;
}

void MM_delete_string(char *s)
{
    free(s);
}

void point_Phi(point_t PhiP, point_t P)
{
    fp2_mul(PhiP->x, P->x, zeta);
    fp2_set(PhiP->y, P->y);
}

static void initpq()
{
    mpz_init(p1onq);
    mpz_add_ui(p1onq, params.p, 1);
    mpz_divexact(p1onq, p1onq, params.q);

    //initialize the Weil pairing library
    curve_init_prime(params.p, params.q);

    fp2_init(zeta);
    fp2_set_cbrt_unity(zeta);
}

static void make_order_q(point_t P)
//P = ((p + 1)/q)P 
//so it is order q
//(except in rare cases)
{
    point_mul(P, p1onq, P);

    /*
    Point check;
    general_group_times(check, p1onq, P);
    if (Q != check) {
	cerr << "GT BUG!\n";
	exit(1);
    }
    */

    //if it's order p1onq, exit
    //XXX: pick new point instead of exiting?
    if (P->infinity) {
	fprintf(stderr, "P can't be made order q!\n");
	exit(1);
    }
}

static void mympz_from_hash(mpz_t x, mpz_t limit,
	const unsigned char *c, const int len)
//let z = number represented in c
//x = z || 1 || z || 2 || z || 3 ...
//until there are enough bits
//should be rewritten more efficiently?
{
    mpz_t z;
    int countlen;
    mpz_t count;
    int i = 0, j;
    int bits = mpz_sizeinbase(limit, 2);
    int zbits;

    mpz_init(z);
    mpz_init(count);
    mpz_set_ui(count, 1);

    mympz_inp_raw(z, c, len);
    zbits = mpz_sizeinbase(z, 2);

    //printf("bits: " << bits << endl;
    mpz_set_ui(x, 0);
    do {
	for (j=0; j<zbits; j++) {
	    if (mpz_tstbit(z, j)) {
		mpz_setbit(x, i);
	    }
	    i++;
	}
	bits -= zbits;
	//printf("bits: " << bits << endl;
	if (bits <= 0) break;

	countlen = mpz_sizeinbase(count, 2);
	for (j=0; j<countlen; j++) {
	    if (mpz_tstbit(count, j)) {
		mpz_setbit(x, i);
	    }
	    i++;
	}
	bits -= countlen;
	//printf("bits: " << bits << endl;
	mpz_add_ui(count, count, 1);
    } while (bits > 0);
    /*
    printf("hash: ");
    mpz_out_str(NULL, 0, x);
    printf("\n");
    */
    while (mpz_cmp(x, limit) > 0) {
	mpz_clrbit(x, mpz_sizeinbase(x, 2) - 1);
    }

    mpz_clear(z);
    mpz_clear(count);
}

void hash_G(mpz_t h, byte_string_t bs)
{
    unsigned int md_len;
    unsigned char md_value[EVP_MAX_MD_SIZE];

    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, bs->data, bs->len);
    EVP_DigestFinal(&mdctx, md_value, &md_len);

    mympz_from_hash(h, params.p, md_value, md_len);
}

void hash_H(byte_string_t md_value, fp2_t x)
{
    unsigned int md_len;
    byte_string_t bs;

    int na = mympz_sizeinbytes(x->a);
    int nb = mympz_sizeinbytes(x->b);

    byte_string_init(bs, na > nb ? na : nb);

    mympz_out_raw(x->a, bs->data, na);

    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, bs->data, na);

    mympz_out_raw(x->a, bs->data, nb);
    EVP_DigestUpdate(&mdctx, bs->data, nb);

    EVP_DigestFinal(&mdctx, md_value->data, &md_len);
    //assert(md_len == params.n);

    byte_string_clear(bs);
}


void hash_G1(byte_string_t md_value, byte_string_t input)
{
    unsigned int md_len;

    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, input->data, input->len);
    EVP_DigestFinal(&mdctx, md_value->data, &md_len);
    //assert(md_len == params.n);
}

void hash_H1(mpz_t r, byte_string_t c1, byte_string_t c2)
{
    int n = params.n;
    byte_string_t md_value;
    byte_string_init(md_value, n);

    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, c1->data, c1->len);
    EVP_DigestUpdate(&mdctx, c2->data, c2->len);
    EVP_DigestFinal(&mdctx, md_value->data, &md_value->len);

    mympz_from_hash(r, params.q, md_value->data, md_value->len);
    byte_string_clear(md_value);
}

//converts id into a point of order q on E/F_p
static void map_to_point(point_t d, const char *id)
{
    fp2_t x, y;
    byte_string_t bsid;

#ifdef BENCHMARK
    bm_put(bm_get_time(), "mtp0");
#endif

    fp2_init(x);
    fp2_init(y);

    //first use G function to choose x coordinate
    byte_string_init_set(bsid, id);
    hash_G(y->a, bsid);
    byte_string_clear(bsid);

    mpz_set_ui(y->b, 0);

    x_from_y(x->a, y->a);
    mpz_set_ui(x->b, 0);

    //we now have a point d
    fp2_set(d->x, x);
    fp2_set(d->y, y);

#ifdef BENCHMARK
    bm_put(bm_get_time(), "mtp1");
#endif
    //make sure it's of order q
    //modify it if necessary
    make_order_q(d);
#ifdef BENCHMARK
    bm_put(bm_get_time(), "mtp2");
#endif
    /*
    printf("mapped to ");
    point_out_str(NULL, 0, d);
    printf("\n");
    */
}

void IBE_init(void)
{
    OpenSSL_add_all_algorithms();
    md = EVP_get_digestbyname("SHA1");
    sharecipher = EVP_des_ede3_cbc();
    version_string = MM_new_string(strlen(IBE_VERSION) + 1);
    strcpy(version_string, IBE_VERSION);
}

void IBE_setup(int k, int qk, char *id)
/* generate system parameters
 * k = number of bits in p (should be at least 512)
 * qk = number of bits in q (size of subgroup, 160 is typical)
 * id = system ID
 */
{
    point_t P, Ppub;
    BIGNUM *bnp, *bnq, *bnr, *bnx;
    BN_CTX *bc;

    int solinasa, solinasb;

    int kqk = k - qk - 4; //lose 4 bits since 12 is a 4-bit no.
    unsigned int seed;

    bnp = BN_new();
    bnq = BN_new();
    bnr = BN_new();
    bnx = BN_new();
    bc = BN_CTX_new();

    //find random k-bit prime p such that
    //p = 2 mod 3 and q = (p+1)/12r is prime as well for some r

    //now also want q to be a Solinas prime
    //we use random() to help us find one: should be ok
    RAND_bytes((unsigned char *) &seed, sizeof(int));
    srandom(seed);

    for(;;) {
	//BN_generate_prime(bnq, qk, 0, NULL, NULL, NULL, NULL);
	BN_zero(bnq);
	solinasa = qk - 1;
	BN_set_bit(bnq, solinasa);
	BN_zero(bnr);
	solinasb = random() % qk;
	BN_set_bit(bnr, solinasb);
	if (random() % 2) {
	    BN_add(bnq, bnq, bnr);
	} else {
	    BN_sub(bnq, bnq, bnr);
	    solinasb = -solinasb;
	}
	BN_one(bnr);
	if (random() % 2) {
	    BN_add(bnq, bnq, bnr);
	} else {
	    BN_sub(bnq, bnq, bnr);
	    solinasa = -solinasa;
	}
	if (!BN_is_prime_fasttest(bnq, 40, NULL, bc, NULL, 1)) continue;

	BN_rand(bnr, kqk, 0, 0);
	//p = q * r
	BN_mul(bnp, bnq, bnr, bc);

	//r = (((p << 1) + 1) << 1) << 1 (= 12r)
	BN_lshift1(bnr, bnp);
	BN_add(bnr, bnr, bnp);
	BN_lshift1(bnr, bnr);
	BN_lshift1(bnr, bnr);

	//p = r - 1
	BN_one(bnp);
	BN_sub(bnp, bnr, bnp);
	if (BN_is_prime_fasttest(bnp, 40, NULL, bc, NULL, 1)) break;
    }

    //pick master key x from F_q
    BN_rand_range(bnx, bnq);

    mpz_init(params.p);
    mpz_init(params.q);
    mpz_init(master);
    mympz_set_bn(params.p, bnp);
    mympz_set_bn(params.q, bnq);
    mympz_set_bn(master, bnx);

    BN_free(bnp);
    BN_free(bnq);
    BN_free(bnr);
    BN_CTX_free(bc);

    printf("p q: ");
    mpz_out_str(NULL, 0, params.p);
    printf(" ");
    mpz_out_str(NULL, 0, params.q);
    printf("\n");

    initpq();

    printf("master key: ");
    mpz_out_str(NULL, 0, master);
    printf("\n");

    BN_clear_free(bnx);

    //pick random point P of order q from E/F_p
    point_init(P);
    point_random(P);
    make_order_q(P);

    point_init(Ppub);
    point_mul(Ppub, master, P);
    printf("P: ");
    point_out_str(NULL, 0, P);
    printf("\n");
    printf("Ppub: ");
    point_out_str(NULL, 0, Ppub);
    printf("\n");

    params.n = 20; //n <= EVP_MAX_MD_SIZE;
    point_init(params.P);
    point_set(params.P, P);

    point_mul_preprocess(P);

    point_init(params.Ppub);
    point_set(params.Ppub, Ppub);
    miller_cache_init(cache_Ppmc);
    tate_preprocess(cache_Ppmc, Ppub);

    point_Phi(PhiPpub, Ppub);

    params.id = MM_new_string(strlen(id) + 1);
    strcpy(params.id, id);

    params.sharet = params.sharen = 0;
    params.version = version_string;

    /*
    int qbits = NumBits(q);
    params.iP = MM_new_Point_array(qbits);
    
    params.iP[0] = MM_new_Point();
    *params.iP[0] = P;
    for (int i=1; i<qbits; i++) {
	params.iP[i] = MM_new_Point();
	group_plus(*params.iP[i], *params.iP[i-1], *params.iP[i-1]);
    }
    */

    if (0) {
	point_t dP;
	byte_string_t dbs;

	printf("debug P\n");
	point_out_str(NULL, 0, P);
	printf("\n");
	byte_string_init_set_point(dbs, P);
	byte_string_printf(dbs, " %02X");
	printf("\n");
	point_init(dP);
	point_set_byte_string(dP, dbs);
	printf("debug dP\n");
	point_out_str(NULL, 0, P);

    }
    point_clear(P);
}

void IBE_extract(byte_string_t bs, const char *id)
//for testing purposes
//only works if master key is in memory
//ideally this shouldn't be the case
{
    point_t key;

    point_init(key);

    map_to_point(key, id);

    point_mul(key, master, key);
    byte_string_init_set_point(bs, key);

    point_clear(key);
}

int IBE_load_master_share(byte_string_t mshare)
//load a master share from a byte_string
{
    mpz_init(current_master_share_y);
    current_master_share_i = (int) mshare->data[0];
    mympz_set_data_len(current_master_share_y,
	    &mshare->data[1], mshare->len - 1);
    return 1;
}

void IBE_clear_master_share(byte_string_t mshare)
//free previously loaded master share
{
    mpz_clear(current_master_share_y);
}

int IBE_extract_share(byte_string_t share, const char *id)
//extract a private key share from an ID
//(using previously loaded master share)
{
    byte_string_t bstemp;
    point_t yd;
    point_t d;

    point_init(yd);
    point_init(d);

    map_to_point(d, id);

    point_mul(yd, current_master_share_y, d);

    byte_string_init(share, 1);
    share->data[0] = (unsigned char) current_master_share_i;
    byte_string_init_set_point(bstemp, yd);
    byte_string_cat(share, bstemp);
    byte_string_clear(bstemp);
    point_clear(yd);
    point_clear(d);

    return 1;
}

int IBE_construct_key(byte_string_t key, byte_string_t *kshare)
//reconstruct a key from key shares
{
    int i, j;
    int indexi, indexj;
    point_t yP;
    mpz_t num, denom;
    mpz_t z;
    point_t d;

    point_init(yP);
    mpz_init(z);
    mpz_init(num);
    mpz_init(denom);
    point_init(d);
    point_set_O(d);
    for (i=0; i<params.sharet; i++) {
	mpz_set_ui(num, 1);
	mpz_set_ui(denom, 1);
	for (j=0; j<params.sharet; j++) {
	    if (j != i) {
		indexi = (int) kshare[i]->data[0];
		indexj = (int) kshare[j]->data[0];
		mpz_mul(num, num, params.robustx[indexj]);
		mpz_mod(num, num, params.q);
		mpz_sub(z, params.robustx[indexj], params.robustx[indexi]);
		mpz_mul(denom, denom, z);
		mpz_mod(denom, denom, params.q);
	    }
	}
	mpz_invert(denom, denom, params.q);
	mpz_mul(z, num, denom);
	mpz_mod(z, z, params.q);
	point_set_data_len(yP, &kshare[i]->data[1], kshare[i]->len - 1);
	point_mul(yP, z, yP);
	point_add(d, d, yP);
    }
    byte_string_init_set_point(key, d);

    point_clear(yP);
    mpz_clear(z);
    mpz_clear(num);
    mpz_clear(denom);
    point_clear(d);

    return 1;
}

void IBE_authenticated_hidesigma(byte_string_t U, byte_string_t *Varray,
    byte_string_t sigma, byte_string_t hash,
    char **idarray,
    byte_string_t sender_key)
//encrypt a key sigma (which must be n bits long)
//output in U, Varray
//must be given the hash of the plaintext message,
//and idarray holds the public keys to encrypt for
//requires Varray has same length has idarray
{
    int i, j;
    int count;
    int n = params.n;
    mpz_t r;
    fp2_t s;
    point_t Qid;
    point_t PhixQ;
    byte_string_t md_value;

    byte_string_t tempbs, tempbs2;

    for (count=0; idarray[count]; count++);
    if (count == 0) return;

    //r = H1(sigma, M)
    mpz_init(r);
    hash_H1(r, sigma, hash);
#ifdef BENCHMARK
    bm_put(bm_get_time(), "rP0");
    bm_put(bm_get_time(), "rP1");
#endif
    //no need to work out rP any more!

    //U = r
    byte_string_init_set_mpz(U, r);

    point_init(Qid);

    point_init(PhixQ);
    point_set_byte_string(PhixQ, sender_key);
    point_Phi(PhixQ, PhixQ);

    byte_string_init(md_value, n);

    for (i=0; i<count; i++) {
	const char *id = idarray[i];

	//XXX:set up a cache to avoid these expensive ops
	map_to_point(Qid, id);
	//calculate s = e(Q_id, Phi(xQ))
	tate_pairing(s, Qid, PhixQ);

#ifdef BENCHMARK
	bm_put(bm_get_time(), "gidr0");
#endif
	//second part of cipher text: V = sigma xor H(U, s)

	byte_string_copy(tempbs, U);
	byte_string_init_set_fp2(tempbs2, s);
	byte_string_cat(tempbs, tempbs2);
	hash_G1(md_value, tempbs);
	byte_string_clear(tempbs);

	byte_string_init(Varray[i], n);
	for (j=0; j<n; j++) {
	    Varray[i]->data[j] = md_value->data[j] ^ sigma->data[j];
	}

#ifdef BENCHMARK
	bm_put(bm_get_time(), "gidr1");
#endif
    }
    byte_string_clear(md_value);

    point_clear(Qid);
    mpz_clear(r);
    fp2_clear(s);
}

void IBE_get_shared_secret_preprocess(byte_string_t key)
{
    point_t Q;

    point_init(Q);

    point_set_byte_string(Q, key);
    miller_cache_init(cache_ssmc);
    tate_preprocess(cache_ssmc, Q);

    point_clear(Q);
}

void IBE_get_shared_secret_postprocess(byte_string_t s, char *id)
{
    point_t Qid;
    fp2_t gid;
    int n = params.n;

    fp2_init(gid);
    point_init(Qid);
    map_to_point(Qid, id);
    //calculate gid = e(Q_id, Phi(P_pub))
    point_Phi(Qid, Qid);
    tate_postprocess(gid, cache_ssmc, Qid);

    byte_string_init(s, n);
    hash_H(s, gid);

    fp2_clear(gid);
    point_clear(Qid);
}

void IBE_get_shared_secret(byte_string_t s, char *id, byte_string_t key)
{
    point_t Qid;
    fp2_t gid;
    int n = params.n;
    point_t Q;

    point_init(Q);

    fp2_init(gid);
    point_init(Qid);
    map_to_point(Qid, id);
    //calculate gid = e(Q_id, Phi(P_pub))
    point_set_byte_string(Q, key);
    point_Phi(Qid, Qid);
    tate_pairing(gid, Q, Qid);

    byte_string_init(s, n);
    hash_H(s, gid);

    fp2_clear(gid);
    point_clear(Q);
    point_clear(Qid);
}

void IBE_hidesigma(byte_string_t U, byte_string_t *Varray,
    byte_string_t sigma, byte_string_t hash,
    char **idarray)
//encrypt a key sigma (which must be n bits long)
//output in U, Varray
//must be given the hash of the plaintext message,
//and idarray holds the public keys to encrypt for
//requires Varray has same length has idarray
{
    int i, j;
    int count;
    int n = params.n;
    mpz_t r;
    fp2_t gidr;
    point_t Qid;
    point_t rP;
    byte_string_t md_value;

    for (count=0; idarray[count]; count++);
    if (count == 0) return;

    //r = H1(sigma, M)
    mpz_init(r);
    hash_H1(r, sigma, hash);
#ifdef BENCHMARK
    bm_put(bm_get_time(), "rP0");
#endif

    //U = rP
    point_init(rP);
    //point_mul(rP, r, params.P);
    point_mul_postprocess(rP, r);
#ifdef BENCHMARK
    bm_put(bm_get_time(), "rP1");
#endif
    fp2_init(gidr);

    byte_string_init_set_point(U, rP);

    point_clear(rP);

    point_init(Qid);

    byte_string_init(md_value, n);

    for (i=0; i<count; i++) {
	const char *id = idarray[i];

	//XXX:set up a cache to avoid these expensive ops
	map_to_point(Qid, id);
	point_Phi(Qid, Qid);
	
	//calculate gidr = e(Q_id, Phi(P_pub))^r

	//tate_pairing(gidr, Qid, PhiPpub);
	tate_postprocess(gidr, cache_Ppmc, Qid);

#ifdef BENCHMARK
	bm_put(bm_get_time(), "gidr0");
#endif
	fp2_pow(gidr, gidr, r);
#ifdef BENCHMARK
	bm_put(bm_get_time(), "gidr1");
#endif

	//second part of cipher text = sigma xor H(g_id^r)
	hash_H(md_value, gidr);
	byte_string_init(Varray[i], n);
	for (j=0; j<n; j++) {
	    Varray[i]->data[j] = md_value->data[j] ^ sigma->data[j];
	}
    }
    byte_string_clear(md_value);

    point_clear(Qid);
    mpz_clear(r);
    fp2_clear(gidr);
}

#ifdef BENCHMARK
static void report_bm_encrypt(void)
{
    fprintf(stderr, "benchmarks:\n");
    fprintf(stderr, "%f initialization\n", bm_get("enc0"));
    fprintf(stderr, "%f computing rP\n", bm_get("rP1") - bm_get("rP0"));
    fprintf(stderr, "%f first part of map_to_point\n", bm_get("mtp1") - bm_get("mtp0"));
    fprintf(stderr, "%f make_order_q\n", bm_get("mtp2") - bm_get("mtp1"));
    fprintf(stderr, "%f miller\n", bm_get("miller1") - bm_get("miller0"));
    fprintf(stderr, "%f Tate power\n", bm_get("gidr0") - bm_get("miller1"));
    fprintf(stderr, "%f gid^r\n", bm_get("gidr1") - bm_get("gidr0"));

    fprintf(stderr, "elapsed time: %f\n", bm_get("enc1"));
}
#endif

int IBE_authenticated_encrypt(byte_string_t U, byte_string_t *Varray, byte_string_t W,
	byte_string_t M, char **idarray,
	byte_string_t sender_key)
{
    int n = params.n;
    byte_string_t sigma, G1sigma;
    byte_string_t md_value;
    int saltlen;
    int Wnewlen;
    int outl;
    EVP_CIPHER_CTX ctx;
    unsigned char key[EVP_MAX_KEY_LENGTH];

#ifdef BENCHMARK
    bm_put(bm_get_time(), "enc0");
#endif

    //pick random sigma in {0,1}^n
    byte_string_init(sigma, n);
    RAND_bytes(sigma->data, n);

    //Replace this part if implementing more general encryption function
    //Hash message
    byte_string_init(md_value, n);
    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, M->data, M->len);
    EVP_DigestFinal(&mdctx, md_value->data, &md_value->len);
    IBE_authenticated_hidesigma(U, Varray, sigma, md_value, idarray, sender_key);

    byte_string_clear(md_value);

    saltlen = EVP_CIPHER_iv_length(sharecipher);
    byte_string_init(W, M->len + EVP_CIPHER_block_size(sharecipher) + saltlen);

    //G1(sigma) is the key we use to encrypt
    byte_string_init(G1sigma, n);
    hash_G1(G1sigma, sigma);
    byte_string_clear(sigma);

    //Replace this part if implementing more general encryption function
    //Perform encryption. We tack it on the end of U
    EVP_BytesToKey(sharecipher, EVP_md5(), NULL, G1sigma->data,
	    n, 1, key, NULL);

    byte_string_clear(G1sigma);
    
    if (1 != RAND_bytes(W->data, saltlen)) {
	return 0;
    }

    EVP_EncryptInit(&ctx, sharecipher, key, W->data);
    Wnewlen = saltlen;
    EVP_EncryptUpdate(&ctx, &W->data[Wnewlen], &outl, M->data, M->len);
    Wnewlen += outl;
    EVP_EncryptFinal(&ctx, &W->data[Wnewlen], &outl);
    Wnewlen += outl;
    byte_string_reinit(W, Wnewlen);

#ifdef BENCHMARK
    bm_put(bm_get_time(), "enc1");
#endif

    //uncomment to get benchmarks
#ifdef BENCHMARK
    report_bm_encrypt();
#endif
    return 1;
}

int IBE_encrypt(byte_string_t U, byte_string_t *Varray, byte_string_t W,
	byte_string_t M, char **idarray)
{
    int n = params.n;
    byte_string_t sigma, G1sigma;
    byte_string_t md_value;
    int saltlen;
    int Wnewlen;
    int outl;
    EVP_CIPHER_CTX ctx;
    unsigned char key[EVP_MAX_KEY_LENGTH];

#ifdef BENCHMARK
    bm_put(bm_get_time(), "enc0");
#endif

    //pick random sigma in {0,1}^n
    byte_string_init(sigma, n);
    RAND_bytes(sigma->data, n);

    //Replace this part if implementing more general encryption function
    //Hash message
    byte_string_init(md_value, n);
    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, M->data, M->len);
    EVP_DigestFinal(&mdctx, md_value->data, &md_value->len);
    IBE_hidesigma(U, Varray, sigma, md_value, idarray);

    byte_string_clear(md_value);

    saltlen = EVP_CIPHER_iv_length(sharecipher);
    byte_string_init(W, M->len + EVP_CIPHER_block_size(sharecipher) + saltlen);

    //G1(sigma) is the key we use to encrypt
    byte_string_init(G1sigma, n);
    hash_G1(G1sigma, sigma);
    byte_string_clear(sigma);

    //Replace this part if implementing more general encryption function
    //Perform encryption. We tack it on the end of U
    EVP_BytesToKey(sharecipher, EVP_md5(), NULL, G1sigma->data,
	    n, 1, key, NULL);

    byte_string_clear(G1sigma);
    
    if (1 != RAND_bytes(W->data, saltlen)) {
	return 0;
    }

    EVP_EncryptInit(&ctx, sharecipher, key, W->data);
    Wnewlen = saltlen;
    EVP_EncryptUpdate(&ctx, &W->data[Wnewlen], &outl, M->data, M->len);
    Wnewlen += outl;
    EVP_EncryptFinal(&ctx, &W->data[Wnewlen], &outl);
    Wnewlen += outl;
    byte_string_reinit(W, Wnewlen);

#ifdef BENCHMARK
    bm_put(bm_get_time(), "enc1");

    report_bm_encrypt();
#endif
    return 1;
}

void IBE_authenticated_revealsigma(byte_string_t sigma,
	byte_string_t U, byte_string_t V, byte_string_t key, char *sender)
//reveal a key sigma
//given U, V, and a private key
{
    int i;
    int n = params.n;
    point_t xQ, PhiQ2;
    fp2_t s;
    byte_string_t md_value;
    byte_string_t tempbs, tempbs2;

    byte_string_init(md_value, n);

    byte_string_init(sigma, n);
    fp2_init(s);

    point_init(xQ);
    point_init(PhiQ2);

    point_set_byte_string(xQ, key);

    //XXX:cache this
    //calculate s = e(xQ, Phi(Q2)), where Q2 = public key of sender
    map_to_point(PhiQ2, sender);
    point_Phi(PhiQ2, PhiQ2);
    tate_pairing(s, xQ, PhiQ2);

    //compute sigma = V xor H(U,s)
    byte_string_copy(tempbs, U);
    byte_string_init_set_fp2(tempbs2, s);
    byte_string_cat(tempbs, tempbs2);
    hash_G1(md_value, tempbs);
    byte_string_clear(tempbs);

    for (i=0; i<n; i++) {
	sigma->data[i] = V->data[i] ^ md_value->data[i];
    }
    byte_string_clear(md_value);

    fp2_clear(s);
    point_clear(xQ);
    point_clear(PhiQ2);
}

void IBE_revealsigma_preprocess(byte_string_t key)
//speeds up miller() by caching stuff
{
    point_t xQ;
    point_init(xQ);
    point_set_byte_string(xQ, key);
    miller_cache_init(cache_xQmc);
    tate_preprocess(cache_xQmc, xQ);
}

void IBE_revealsigma_postprocess(byte_string_t sigma,
	byte_string_t U, byte_string_t V)
//reveal a key sigma
//given U, V, and a private key
{
    int i;
    int n = params.n;
    point_t rP;
    fp2_t res;
    byte_string_t md_value;

    byte_string_init(sigma, n);
    fp2_init(res);

    point_init(rP);

    point_set_byte_string(rP, U);

    //Compute sigma = V xor H(e(Pkey, Phi(U)))
    point_Phi(rP, rP);
    tate_postprocess(res, cache_xQmc, rP);
    byte_string_init(md_value, n);
    hash_H(md_value, res);
    for (i=0; i<n; i++) {
	sigma->data[i] = V->data[i] ^ md_value->data[i];
    }
    byte_string_clear(md_value);

    fp2_clear(res);
    point_clear(rP);
}

void IBE_revealsigma(byte_string_t sigma,
	byte_string_t U, byte_string_t V, byte_string_t key)
//reveal a key sigma
//given U, V, and a private key
{
    int i;
    int n = params.n;
    point_t xQ, rP;
    fp2_t res;
    byte_string_t md_value;

    byte_string_init(sigma, n);
    fp2_init(res);

    point_init(xQ);
    point_init(rP);
    point_set_byte_string(xQ, key);
    point_set_byte_string(rP, U);

    //Compute sigma = V xor H(e(Pkey, Phi(U)))
    point_Phi(rP, rP);

    tate_pairing(res, xQ, rP);

    byte_string_init(md_value, n);

    hash_H(md_value, res);
    for (i=0; i<n; i++) {
	sigma->data[i] = V->data[i] ^ md_value->data[i];
    }
    byte_string_clear(md_value);

    fp2_clear(res);
    point_clear(xQ);
    point_clear(rP);
}

int IBE_verify(byte_string_t U, byte_string_t sigma, byte_string_t hash)
//verify that the ciphertext is valid
//given U, a key sigma, and the hash of the plaintext
{
    //Set r = H1(sigma, M)
    mpz_t r;
    point_t rP, allegedrP;

    mpz_init(r);
    hash_H1(r, sigma, hash);

    point_init(rP);
    //point_mul(rP, r, params.P);
    point_mul_postprocess(rP, r);
    mpz_clear(r);

    point_init(allegedrP);
    point_set_byte_string(allegedrP, U);

    if (!point_equal(rP, allegedrP)) {
	printf("bad ciphertext: rP != U\n");
	//printf("rP = ");
	//point_out_str(NULL, 0, rP);
	//printf("\n");
	//printf("U = ");
	//point_out_str(NULL, 0, allegedrP);
	//printf("\n");
	point_clear(rP);
	point_clear(allegedrP);
	return 0;
    }
    point_clear(rP);
    point_clear(allegedrP);
    return 1;
}

int IBE_authenticated_decrypt(byte_string_t M, byte_string_t U, byte_string_t V,
	byte_string_t W, byte_string_t d, char *sender)
{
    int n = params.n;
    byte_string_t sigma, G1sigma;
    int mlen;
    int outl;
    byte_string_t md_value;
    unsigned int md_len;
    int saltlen;

    EVP_CIPHER_CTX ctx;
    unsigned char key[EVP_MAX_KEY_LENGTH];

    IBE_authenticated_revealsigma(sigma, U, V, d, sender);

    //Replace this part if implementing more general decryption function
    //Decrypt message with G1sigma as key

    byte_string_init(G1sigma, n);
    hash_G1(G1sigma, sigma);
    EVP_BytesToKey(sharecipher, EVP_md5(), NULL, G1sigma->data,
	    n, 1, key, NULL);
    byte_string_clear(G1sigma);
    
    saltlen = EVP_CIPHER_iv_length(sharecipher);

    byte_string_init(M, W->len - saltlen + EVP_CIPHER_block_size(sharecipher));

    EVP_DecryptInit(&ctx, sharecipher, key, W->data);
    EVP_DecryptUpdate(&ctx, M->data, &mlen, &W->data[saltlen], W->len - saltlen);
    EVP_DecryptFinal(&ctx, &M->data[mlen], &outl);
    mlen += outl;
    byte_string_reinit(M, mlen);

    //Replace this part if implementing more general decryption function
    //Hash message (for verification purposes)
    byte_string_init(md_value, n);
    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, M->data, mlen);
    EVP_DigestFinal(&mdctx, md_value->data, &md_len);
    //check md_len = md_value->len

    if (!IBE_verify(U, sigma, md_value)) {
	byte_string_clear(md_value);
	byte_string_clear(sigma);
	printf("Message not valid!\n");
	return 0;
    }

    byte_string_clear(md_value);
    byte_string_clear(sigma);
    return 1;
}

int IBE_decrypt(byte_string_t M, byte_string_t U, byte_string_t V,
	byte_string_t W, byte_string_t d)
{
    int n = params.n;
    byte_string_t sigma, G1sigma;
    int mlen;
    int outl;
    byte_string_t md_value;
    unsigned int md_len;
    int saltlen;

    EVP_CIPHER_CTX ctx;
    unsigned char key[EVP_MAX_KEY_LENGTH];

#ifdef BENCHMARK
    bm_put(bm_get_time(), "dec0");
#endif

    IBE_revealsigma(sigma, U, V, d);

    //Replace this part if implementing more general decryption function
    //Decrypt message with G1sigma as key
    byte_string_init(G1sigma, n);

    hash_G1(G1sigma, sigma);
    EVP_BytesToKey(sharecipher, EVP_md5(), NULL, G1sigma->data,
	    n, 1, key, NULL);

    byte_string_clear(G1sigma);
    saltlen = EVP_CIPHER_iv_length(sharecipher);
    byte_string_init(M, W->len - saltlen + EVP_CIPHER_block_size(sharecipher));

    EVP_DecryptInit(&ctx, sharecipher, key, W->data);
    EVP_DecryptUpdate(&ctx, M->data, &mlen, &W->data[saltlen], W->len - saltlen);
    EVP_DecryptFinal(&ctx, &M->data[mlen], &outl);
    mlen += outl;
    byte_string_reinit(M, mlen);

    //Replace this part if implementing more general decryption function
    //Hash message (for verification purposes)
    byte_string_init(md_value, n);
    EVP_DigestInit(&mdctx, md);
    EVP_DigestUpdate(&mdctx, M->data, mlen);
    EVP_DigestFinal(&mdctx, md_value->data, &md_len);
    //check md_len = md_value->len

    if (!IBE_verify(U, sigma, md_value)) {
	byte_string_clear(md_value);
	byte_string_clear(sigma);
	printf("Message not valid!\n");
	return 0;
    }

    byte_string_clear(md_value);
    byte_string_clear(sigma);

#ifdef BENCHMARK
    bm_put(bm_get_time(), "dec1");

    fprintf(stderr, "dec time: %f\n", bm_get("dec1") - bm_get("dec0"));
#endif

    return 1;
}

char *IBE_get_year(void)
{
    time_t tt;
    struct tm *tmt;
    char *s;
    int year;

    time(&tt);
    tmt = gmtime(&tt);
    s = MM_new_string(5);

    year = 1900 + tmt->tm_year;
    if (year > 9999) year = 9999;
    //Y10K bug: need to fix eventually :)

    sprintf(s, "%d", year);
    return s;
}

char *IBE_make_id(char *addr, char *subject)
{
    char *year = IBE_get_year();
    int i;
    char *s;

    if (!addr) {
	return NULL;
    }

    //XXX: yet another magic number
    i = strlen(addr) + strlen(params.id) + 1000;

    if (subject) i += strlen(subject);
    s = MM_new_string(i);

    i = sprintf(s, "purpose=email&date=%s&id=%s&", year, addr);
    if (subject) {
	i += sprintf(&s[i], "subject=%s&", subject);
    }
    sprintf(&s[i], "system=%s", params.id);

    MM_delete_string(year);
    return s;
}

char* IBE_version(void)
{
    return params.version;
}

int IBE_threshold(void)
{
    return params.sharet;
}

void IBE_split_master(byte_string_t *mshare, int t, int n)
//split the master key into n pieces
//t of them are required to generate a key
{
    mpz_t poly[t];
    mpz_t *x;
    mpz_t y[n];
    mpz_t z;
    BIGNUM *bnr, *bnq;
    int i, j;
    byte_string_t bstemp;
    //start with random polynomial with degree <= t-2
    //(in F_q)
    //whose constant term = master

    bnr = BN_new();
    bnq = BN_new();
    mympz_get_bn(params.q, bnq);

    mpz_init(poly[0]);
    mpz_set(poly[0], master);
    for (i=1; i<t; i++) {
	mpz_init(poly[i]);
	BN_rand_range(bnr, bnq);
	mympz_set_bn(poly[i], bnr);
    }

    mpz_init(z);

    printf("secret poly: ");
    for (i=0; i<t; i++) {
	printf(" ");
	mpz_out_str(NULL, 0, poly[i]);
    }
    printf("\n");

    params.robustx = (mpz_t *) malloc(n * sizeof(mpz_t));
    params.robustP = (point_t *) malloc(n * sizeof(point_t));
    params.sharet = t;
    params.sharen = n;
    x = params.robustx;

    for (i=0; i<n; i++) {
	BN_rand_range(bnr, bnq);
	mpz_init(x[i]);
	mympz_set_bn(x[i], bnr);
	//Horner's rule
	mpz_set_ui(z, 0);
	for (j=t-1; j>=0; j--) {
	    mpz_mul(z, z, x[i]);
	    mpz_add(z, z, poly[j]);
	    mpz_mod(z, z, params.q);
	}
	mpz_init_set(y[i], z);
	byte_string_init(mshare[i], 1);
	mshare[i]->data[0] = (unsigned char) i;
	byte_string_init_set_mpz(bstemp, z);
	byte_string_cat(mshare[i], bstemp);
    }

    for (i=0; i<n; i++) {
	point_init(params.robustP[i]);
	point_mul(params.robustP[i], y[i], params.Ppub);
    }

    if (0) { //for testing, has memory leaks
    point_t myid;
    point_t P[t];
    point_t Q, R;
    mpz_t num, denom;
    int RisO = 1;

    printf("TESTING\n");

    point_init(myid);
    map_to_point(myid, "blynn@cs.stanford.edu");

    printf("myid = ");
    point_out_str(NULL, 0, myid);
    printf("\n");

    for (i=0; i<t; i++) {
	point_init(P[i]);
	point_mul(P[i], y[i], myid);
	printf("P[i] = ");
	point_out_str(NULL, 0, P[i]);
	printf(", y[i] = ");
	mpz_out_str(NULL, 0, y[i]);
	printf("\n");
    }

    mpz_init(num);
    mpz_init(denom);
    point_init(Q);
    point_init(R);
    for (i=0; i<t; i++) {
	mpz_set_ui(num, 1);
	mpz_set_ui(denom, 1);
	for (j=0; j<t; j++) {
	    if (j != i) {
		mpz_mul(num, num, params.robustx[j]);
		mpz_mod(num, num, params.q);
		mpz_sub(z, params.robustx[j], params.robustx[i]);
		mpz_mul(denom, denom, z);
		mpz_mod(denom, denom, params.q);
	    }
	}
	mpz_invert(denom, denom, params.q);
	mpz_mul(z, num, denom);
	mpz_mod(z, z, params.q);
	point_mul(Q, z, P[i]);
	if (RisO) {
	    point_set(R, Q);
	    RisO = 0;
	} else {
	    point_add(R, R, Q);
	}
    }
	
    printf("R = ");
    point_out_str(NULL, 0, R);
    printf("\n");

    point_mul(R, master, myid);
    printf("R = ");
    point_out_str(NULL, 0, R);
    printf("\n");

    }

    for (i=0; i<t; i++) {
	mpz_clear(poly[i]);
    }

    for (i=0; i<n; i++) {
	mpz_clear(y[i]);
    }

    mpz_clear(z);

    BN_free(bnq);
    BN_clear_free(bnr);
}

int IBE_construct_master(byte_string_t *mshare, int t)
//reconstruct the master from master shares
//(shouldn't normally be used since they should never be in one place)
{
    int i, j;
    int indexi, indexj;
    mpz_t y;
    mpz_t num, denom;
    mpz_t z;

    mpz_init(y);
    mpz_init(z);
    mpz_init(num);
    mpz_init(denom);
    mpz_init(master);
    mpz_set_ui(master, 0);
    for (i=0; i<t; i++) {
	mpz_set_ui(num, 1);
	mpz_set_ui(denom, 1);
	for (j=0; j<t; j++) {
	    if (j != i) {
		indexi = (int) mshare[i]->data[0];
		indexj = (int) mshare[j]->data[0];
		mpz_mul(num, num, params.robustx[indexj]);
		mpz_mod(num, num, params.q);
		mpz_sub(z, params.robustx[indexj], params.robustx[indexi]);
		mpz_mul(denom, denom, z);
		mpz_mod(denom, denom, params.q);
	    }
	}
	mpz_invert(denom, denom, params.q);
	mpz_mul(z, num, denom);
	mpz_mod(z, z, params.q);
	mympz_set_data_len(y, &mshare[i]->data[1], mshare[i]->len - 1);
	mpz_mul(z, z, y);
	mpz_mod(z, z, params.q);
	mpz_add(master, master, z);
	if (mpz_cmp(master, params.q) >= 0) {
	    mpz_sub(master, master, params.q);
	}
    }

    mpz_clear(y);
    mpz_clear(z);
    mpz_clear(num);
    mpz_clear(denom);

    return 1;
}

//a bit hackish
static void serialize_int(byte_string_t bs, int n)
{
    byte_string_init(bs, 2);
    bs->data[0] = (unsigned char) n >> 8;
    bs->data[1] = (unsigned char) n;
}

static int deserialize_int(unsigned char *data, int *n)
{
    *n = (data[0] << 8) + data[1];
    return 2;
}

static void serialize_string(byte_string_t bs, char *s)
{
    byte_string_t bs2;
    int n = strlen(s);

    serialize_int(bs, n);
    byte_string_init_set(bs2, s);
    byte_string_cat(bs, bs2);
}

static int deserialize_string(unsigned char *data, char **s)
{
    int n;
    deserialize_int(data, &n);
    *s = (char *) malloc(n + 1);
    memcpy(*s, &data[2], n);
    (*s)[n] = 0;
    return n + 2;
}

int IBE_serialize_params(byte_string_t bs)
//put system parameters into a byte_string
{
    int i;
    byte_string_t bs2;

    serialize_string(bs, params.version);
    serialize_string(bs2, params.id);
    byte_string_cat(bs, bs2);
    byte_string_init_set_mpz(bs2, params.p);
    byte_string_cat(bs, bs2);
    byte_string_init_set_mpz(bs2, params.q);
    byte_string_cat(bs, bs2);
    serialize_int(bs2, params.n);
    byte_string_cat(bs, bs2);
    byte_string_init_set_point(bs2, params.P);
    byte_string_cat(bs, bs2);
    byte_string_init_set_point(bs2, params.Ppub);
    byte_string_cat(bs, bs2);
    serialize_int(bs2, params.sharet);
    byte_string_cat(bs, bs2);
    serialize_int(bs2, params.sharen);
    byte_string_cat(bs, bs2);

    for (i=0; i<params.sharen; i++) {
	byte_string_init_set_mpz(bs2, params.robustx[i]);
	byte_string_cat(bs, bs2);
	byte_string_init_set_point(bs2, params.robustP[i]);
	byte_string_cat(bs, bs2);
    }

    return 1;
}

int IBE_deserialize_params(byte_string_t bs)
//get system parameters from a byte_string
{
    int i;
    int offset;
    int len = bs->len;
    unsigned char *data = bs->data;

    offset = deserialize_string(data, &params.version);
    offset += deserialize_string(&data[offset], &params.id);

    mpz_init(params.p);
    offset += mympz_set_data_len(params.p, &data[offset], len - offset);

    mpz_init(params.q);
    offset += mympz_set_data_len(params.q, &data[offset], len - offset);

    initpq();

    offset += deserialize_int(&data[offset], &params.n);

    point_init(params.P);
    offset += point_set_data_len(params.P, &data[offset], len - offset);

    point_init(params.Ppub);
    offset += point_set_data_len(params.Ppub, &data[offset], len - offset);

    point_Phi(PhiPpub, params.Ppub);

    offset += deserialize_int(&data[offset], &params.sharet);
    offset += deserialize_int(&data[offset], &params.sharen);

    params.robustx = (mpz_t *) malloc(params.sharen * sizeof(mpz_t));
    params.robustP = (point_t *) malloc(params.sharen * sizeof(point_t));
    for (i=0; i<params.sharen; i++) {
	mpz_init(params.robustx[i]);
	offset += mympz_set_data_len(params.robustx[i], &data[offset], len - offset);
	point_init(params.robustP[i]);
	offset += point_set_data_len(params.robustP[i], &data[offset], len - offset);
    }

    point_mul_preprocess(params.P);
    miller_cache_init(cache_Ppmc);
    tate_preprocess(cache_Ppmc, params.Ppub);

    return 1;
}

void IBE_print_params(void)
{
    printf("IBE version: %s\n", params.version);
    printf("system ID: %s\n", params.id);

    printf("p = ");
    mpz_out_str(NULL, 0, params.p);
    printf("\n");
    printf("q = ");
    mpz_out_str(NULL, 0, params.q);
    printf("\n");
    printf("P = ");
    point_out_str(NULL, 0, params.P);
    printf("\n");
    printf("Ppub = ");
    point_out_str(NULL, 0, params.Ppub);
    printf("\n");
    printf("n = %d\n", params.n);
    printf("share t/n = %d/%d\n", params.sharet, params.sharen);
}
