    #include <stdio.h>
    #include <stdlib.h>
    #include <unistd.h>
    #include <errno.h>
    #include <string.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <sys/wait.h>
    #include <signal.h>
    #include "ibe.h"
    #include "format.h"
    #include "ibe_progs.h"

    #define MYPORT 3490        // the port users will be connecting to
    #define MAXDATASIZE 1000   // the maximum num bytes to receive
    #define BACKLOG 10         // how many pending connections queue will hold

    CONF_CTX *cnfctx;

    /* this function kills off zombies that occur from fork()'ed processes */ 
    void sigchld_handler(int s)
    {
        while(wait(NULL) > 0);
    }


    int main(int argc, char **argv)
    {
        int sockfd, new_fd;  // listen on sock_fd, new connection on new_fd
        struct sockaddr_in my_addr;    // my address information
        struct sockaddr_in their_addr; // connector's address information
        socklen_t sin_size;
        struct sigaction sa;
        int yes = 1;
	int textlen;
	FILE *fp;
	int portnum;
	int numbytes;

	byte_string_t key;
	byte_string_t M;
	char *pw;
	char *privkeyfile;
	char *paramsfile;
	int status;
	char cnffile[] = "server.cnf";
	char tmpfile[] = "temp.svr";
        char buf[MAXDATASIZE];

	IBE_init();	

	printf("\nIBE-SSL server v1.0\nby J. Adam Sowers (jasowers@virginia.edu)\n");
	printf("based on Stanford IBE 0.21\n\n");

	// load configuration file
	cnfctx = LoadConfig(cnffile);
	if (!cnfctx) {
		fprintf(stderr, "error opening %s\n", cnffile);
		fprintf(stderr, "using default values\n");
		cnfctx = constructCTX();
	}

	paramsfile = GetPathParam(cnfctx, "params", 0, "params.txt");
	status = FMT_load_params(paramsfile);
	if (status != 1) {
		fprintf(stderr, "error loading params file %s\n", paramsfile);
		return(1);
	}

	// get parameters from configuration file
	portnum = GetIntParam(cnfctx, "port", 0, 31832);
	pw = GetStringParam(cnfctx, "password", 0, "");
	privkeyfile = GetPathParam(cnfctx, "path", 0, "keyfile");

	// set up socket
        if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
            perror("socket");
            exit(1);
        }
        if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) 
	{
            perror("setsockopt");
            exit(1);
        }
        
	// standard TCP/IP socket parameters
        my_addr.sin_family = AF_INET;         // host byte order
        my_addr.sin_port = htons(portnum);     // short, network byte order
        my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
        memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct

	// bind socket
        if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) 
	{
            perror("bind");
            exit(1);
        }

	// listen on socket
        if (listen(sockfd, BACKLOG) == -1) 
	{
            perror("listen");
            exit(1);
        }

        sa.sa_handler = sigchld_handler; // reap all dead processes
        sigemptyset(&sa.sa_mask);
        sa.sa_flags = SA_RESTART;
        if (sigaction(SIGCHLD, &sa, NULL) == -1) 
	{
            perror("sigaction");
            exit(1);
        }
	
	// main accept() loop
        while(1) 
        {
            sin_size = sizeof(struct sockaddr_in);
            if ((new_fd = accept(sockfd, (struct sockaddr *)&their_addr,
                                                           &sin_size)) == -1) 
	    {
                perror("accept");
                continue;
            }
 
            // received a new connection
            printf("server: got connection from %s\n", inet_ntoa(their_addr.sin_addr));
 
 	    // test for child process
            if (!fork()) 
            {
                close(sockfd); // child doesn't need the listener
                
                if (send(new_fd, "IBE-SSL server v1.0\r\n", 21, 0) == -1)
                    perror("send");

		// open a temporary file for storing the encrypted text
		fp = fopen(tmpfile, "w");

		printf("Receving encrypted message...\n");

		// store incoming message to the temp file
		numbytes = 1;
		while(numbytes) 
		{
			if ((numbytes = recv(new_fd, buf, MAXDATASIZE - 1, 0)) == -1) 
			{
			    perror("recv");
        		    exit(1);
       		 	}

	 	    buf[numbytes] = '\0';
			fprintf(fp, "%s", buf);
			
			textlen = 0;
			if(strcmp(buf, "-----END IBE-----\n") == 0) numbytes = 0;
		}

		fflush(fp);
		fclose(fp);

		printf("Finished receiving message.\n");

		// reopen file to load text for decryption
		fp = fopen(tmpfile, "r");

		// load the private key file
		printf("Loading private key file...\n");
		status = FMT_crypt_load(privkeyfile, key, pw);	
		if (status != 1) 
		{
			fprintf(stderr, "error loading private key %s\n", privkeyfile);
			return 1;
		}

		// decrypt using IBE functions
		printf("Decrypting message...\n");
		status = FMT_decrypt(M, fp, key);
		if (status != 1) 
		{
			fprintf(stderr, "error in decryption\n");
			return 1;
		}

		printf("Message decrypted.\n");

		fclose(fp);
		remove(tmpfile);
		
		// open the file again, this time to write out the decrypted message
		fp = fopen(tmpfile, "w");
		byte_string_fprintf(fp, M, "%c");
		fclose(fp);
		
		// open once more, to send the message back to the client
		fp = fopen(tmpfile, "r");
		printf("Sending unencrypted message...\n");
		fgets(buf, 256, fp);
		while(!feof(fp))
		{
			send(new_fd, buf, 256, 0);
			fgets(buf, 256, fp);
		}

		// send disconnection notice to client
		send(new_fd, "!!!disconnect!!!", 16, 0);

		// remove temporary file
		remove(tmpfile);
		
		// close up connection
		printf("Finished sending unencrypted message. Closing connection...\n");
		close(new_fd);
                exit(0);	
            }

            close(new_fd);  // parent doesn't need this
        }

        return 0;
    } 

